univariateDensityAdaptive_Parallel <- function(u, ngrid=300,weightsObs=rep(1,length(u)),alpha=0.5,
          rangeObs=c((min(u - 3*weighted.sd(u,weightsObs))),(max(u + 3*weighted.sd(u,weightsObs)))) ,
          figure=TRUE,
          method.h="normal",
          provide.h=FALSE,
          h.u.pilot= (4/3)^(1/5)*n^(-1/5)*sum(((u-weighted.mean(u,weightsObs/sum(weightsObs)))^2)*weightsObs/sum(weightsObs))^(1/2),
          provide.lambda=FALSE,
          lambdas.provide=rep(1,length(u)),
          provide.eval.points=FALSE,
          eval.points.provide=rep(NA,length=ngrid),
          logTrans=FALSE,
          kernelType="Gaussian",
          addFigure=FALSE,
          colProvided="black",
          lwdProvided=2,
          xlab="Variable",
          xlim= rangeObs){

##Univariate density estimation by adaptive kernel
##See Silverman (1986), pp. 100-110


  # u: vector of observations
  # ngrid: number of points in the grid
  # alpha: parameter of the adaptive kernel (default=0.5). Alpha=0 no adaptive kernel
  # weightsObs: weights of observations
  
  # method.h: method to compute the bandwidth ("normal" (default), "sj", "cv", "aicc"). See documentation of "sm" package
  # h.u.pilot: bandwidth of pilot estimate
  # provide.h: bandwidth is passed to the routine
  # provide.lambda: lambdas are passed to the routine and not calculated by pilot estimate

  # provide.eval.points: evaluation.points are passed to the routine (rangeObs is then not considered)

  #logTrans: log transformation of data

  #kernelType: "Gaussian"(default) or "Epanechnikov" or "uniform"
  
  ###########################################################################
  ## Gaussian, Epanechnikov, and uniform univariate kernels for the estimates Silverman (1986) p. 43
  
    normKernel <- function(eval.point,u,h.u,lambdas,weightsObsNorm){
		    			return(sum((1/(2*pi)^(1/2))*exp(-(1/2)*((u-eval.point)/(lambdas*h.u))^2)*weightsObsNorm/(lambdas*h.u)))
                  }

    epaKernel <- function(eval.point,u,h.u,lambdas,weightsObsNorm){
                       K <- (3/4)*(1-((u-eval.point)/(lambdas*h.u))^2)*weightsObsNorm/(lambdas*h.u)
                       ppp <- abs((u-eval.point)/(lambdas*h.u)) >= 1
                       K[ppp] <- 0 
                       return(sum(K))
                   } 

     uniKernel <- function(eval.point,u,h.u,lambdas,weightsObsNorm){
		      		 n <- length(u)
                       K <- c(matrix(1/2,1,n))*weightsObsNorm/(lambdas*h.u)
                       ppp <- abs((u-eval.point)/(lambdas*h.u)) >= 1
                       K[ppp] <- 0 
                       return(sum(K))
                   } 

                   
  ##Log transformation of observations
    if(logTrans==TRUE){
      u <- log(u)
    }

  ##Normalization of weights
    weightsObsNorm <- weightsObs/sum(weightsObs)

  ##Number of observations
    n <- length(u)

  ##Normal bandwidth for the pilot estimates
    if(provide.h==FALSE){

      if (method.h=="normal"){
        sigma.wei <- sum(((u-weighted.mean(u,weightsObsNorm))^2)*weightsObsNorm)^(1/2)
        ##See Silverman (1986), p. 45
        h.u <- (4/3)^(1/5)*n^(-1/5)*sigma.wei
      }else{
        print("Still to be implemented!")
      }

    }else{
      h.u <- h.u.pilot
    }

  ##Evaluation points
    if(provide.eval.points==FALSE){

       eval.points <- seq((min(u - 3*weighted.sd(u,weightsObsNorm))),(max(u + 3*weighted.sd(u,weightsObsNorm))),length=ngrid)

    }else{

       eval.points <- eval.points.provide
       if(logTrans==TRUE){     
	eval.points <- log(eval.points)
       }

    }
   
  ##Pilot estimate, p. 101 Silverman (1986)
  
  ##Gaussian kernel
  if(kernelType=="Gaussian"){

     if(provide.lambda==FALSE){
	
		densityPilotEstimate <- foreach(i=1:n,.combine=c,.verbose=FALSE) %dopar%{  
	  		normKernel(eval.point=u[i],u=u,h.u=h.u,lambdas=1,weightsObsNorm=weightsObsNorm)
	    }
	
        lambdas <- (densityPilotEstimate/exp(sum(log(densityPilotEstimate)*weightsObsNorm)))^(-alpha)
        
      }else{
        lambdas <- lambdas.provide
      }

      ##The estimate of density
      estimate <- foreach(j=1:ngrid,.combine=c,.verbose=FALSE) %dopar%{  
		  normKernel(eval.point=eval.points[j],u=u,h.u=h.u,lambdas=lambdas,weightsObsNorm=weightsObsNorm)
      }
	
    }
  
  ## Epanechnikov kernel
  if(kernelType=="Epanechnikov"){

      if(provide.lambda==FALSE){
    
        densityPilotEstimate <- foreach(i=1:n,.combine=c,.verbose=FALSE) %dopar%{  
	  		epaKernel(eval.point=u[i],u=u,h.u=h.u,lambdas=1,weightsObsNorm=weightsObsNorm)
	    }
	
        lambdas <- (densityPilotEstimate/exp(sum(log(densityPilotEstimate)*weightsObsNorm)))^(-alpha)

      }else{
        lambdas <- lambdas.provide
      }

      ##The estimate of density
       estimate <- foreach(j=1:ngrid,.combine=c,.verbose=FALSE) %dopar%{  
		  epaKernel(eval.point=eval.points[j],u=u,h.u=h.u,lambdas=lambdas,weightsObsNorm=weightsObsNorm)
       }
	
     }

  ## Uniform kernel   
  if(kernelType=="uniform"){

      if(provide.lambda==FALSE){
      	densityPilotEstimate <- foreach(i=1:n,.combine=c,.verbose=FALSE) %dopar%{  
	  		uniKernel(eval.point=u[i],u=u,h.u=h.u,lambdas=1,weightsObsNorm=weightsObsNorm)
	    }
	
        lambdas <- (densityPilotEstimate/exp(sum(log(densityPilotEstimate)*weightsObsNorm)))^(-alpha)
      }else{
        lambdas <- lambdas.provide
      }

      ##The estimate of density
      estimate <- foreach(j=1:ngrid,.combine=c,.verbose=FALSE) %dopar%{  
		  uniKernel(eval.point=eval.points[j],u=u,h.u=h.u,lambdas=lambdas,weightsObsNorm=weightsObsNorm)
       }
     
     }

  ##If observations are log-transformed for the estimate
  if(logTrans==TRUE){
    u <- exp(u)
    eval.points <- exp(eval.points)
    rangeObs <- c(min(eval.points),max(eval.points))
    estimate <- estimate/numericalIntegration(estimate,eval.points)$fdensity
  }
  
  if(figure==TRUE){
    
    if(addFigure==FALSE){

      plot(
      	  #eval.points[round(ngrid*0.2):round(ngrid*0.8)],estimate[round(ngrid*0.2):round(ngrid*0.8)],
      	  eval.points,estimate,
           ,xlab=xlab,ylab=paste("Adaptive kernel estimate of density (",kernelType,"kernel )"),type="l",
           ,col=colProvided,lwd=lwdProvided
           ,
           #ylim=c(0,max(estimate[round(ngrid*0.2):round(ngrid*0.8)]))
           ylim=c(0,max(estimate)),
           xlim=xlim
           )
           
           points(u,rep(0,length=length(u)),pch=".",cex=3)
           
    }else{
      lines(eval.points[round(ngrid*0.05):round(ngrid*1)],estimate[round(ngrid*0.05):round(ngrid*1)],col=colProvided,
      lwd=lwdProvided)
    }
  
  }
  
  list(estimate=estimate,eval.points=eval.points,
   weightsObsNorm=weightsObsNorm,ngrid=ngrid,
   h.u=h.u,lambdas=lambdas,u=u,rangeObs=rangeObs)
 
 }
